//-----------------------------------------
// LScript Generic Update Bones 2
//
// Copyright 2007 Michael Green
// mike@mikegreen.name
//

myversion=1.8;

//v1.8
// Now (optionally) moves IK goals to their chain ends
//v1.7
// work around since 9.6 returns another valid OA instead of 'nil' when an item is deleted! Argh!
//v1.6
// Changed so it restores the current frame.
//v1.5
// Fixed error for deleted bones in MatchBones.
//v1.4
// Fixed error for goals of pivot rotated items. 
//v1.3
// New bones don't have plugins!
//v1.2
// Wasn't actually turning plugins back on :P
//v1.1
// Now uses pivots for goal items, rather than just position.

@version 2.2
@warnings
@script generic

// global values go here
gset;

st_act=1;
st_del=2;
st_rec=3;
st_par=4;
st_fra=5;
st_iku=6;


st_bones=1;
st_flags=2;
st_match=3;
st_paren=4;
st_goals=5;
st_targe=6;
st_pivot=7;
st_mplug=8;

gbones;    // Old bones and properties
gnewbones; // List of clone's bones
gdelbones; // List of bones to be deleted

generic
{
	myitem=Scene().firstSelect();
	if(myitem.isBone())
		myitem=FindMeshParent(myitem);
    if (!myitem.isMesh())
    {
        error("Please select an object with bones.");
        return;
	}
    myoldbone=myitem.bone();
    if (!myoldbone)
    {
        error("Please select an object with bones.");
        return;
	}

    for(i=1;i<=6;i++)
    {
    	gset[i]=recall("MGUpdatebones:"+i.asStr(),1);
    }
	// perform all processing here

    reqbegin("MGUpdate Bones Version:"+myversion);
    c0 = ctltext("Warning!","Ensure AutoKey is ON before running this plugin!!");
    d0 = ctltext("Note","In case of crash a backup file is made in your temp",
						"directory called MGBoneUpdate_Backup.lws");
    c[st_act] = ctlcheckbox("Activate new bones",gset[st_act]);
    c[st_del] = ctlcheckbox("Delete obsolete bones",gset[st_del]);
    c[st_rec] = ctlcheckbox("Record Pivot Rotation",gset[st_rec]);
    c[st_par] = ctlcheckbox("Preserve current parenting",gset[st_par]);
    c[st_iku] = ctlcheckbox("Move IK goals ",gset[st_iku]);
    c[st_fra] = ctlinteger("Setup Frame",gset[st_fra]);

    return if !reqpost();

    // get requester control values here
    for(i=1;i<=gset.size();i++)
    {
		gset[i]=getvalue(c[i]);
    	store("MGUpdatebones:"+i.asStr(),gset[i]);
	}

    reqend();
    
    AutoConfirm(1);

    result=UpdateBones(myitem);

    AutoConfirm(0);

	RefreshNow();
    if(result)
    	info("Bones updated for item:"+myitem.name);
    else
    	error("Bones not updated for item:"+myitem.name);


}

FindMeshParent:mybone
{
	while(mybone.isBone())
	{
		mybone=mybone.parent;
	}
	return(mybone);
}

UpdateBones:pitem
{
	BackupScene();

	oldtime=Scene().currenttime;
    GoToFrame(gset[st_fra]);

    oldpip=Scene().generalopts[3];
	EnableIK();
	EnableMC();

	mynewitem=CreateAClone(pitem);
	if(!mynewitem)
		return(nil);
	PrepareBones(pitem);
	MakeNewBoneList(mynewitem);
	MatchBones();
	MakeNewBones(pitem,newitem);
	PositionBones(pitem);

	if(gset[st_par])
    	ReparentBones();
	if(gset[st_del])
    	DeleteBonesReparent(pitem);

    RestBones();
    RefreshNow();
    
	MoveIKGoals();
	if(gset[st_del])
		ClearBones(gdelbones);


	SelectItem(mynewitem.id);
	ClearSelected();

	EnableIK();
	EnableMC();
	SelectItem(pitem.id);
	GoToFrame(oldtime*Scene().fps);
	return(1);
}

BackupScene
{
    tempdir=getdir(TEMPDIR);
    myscenefile=tempdir+"\\MGBoneUpdate_Backup.lws";
    SaveSceneCopy(myscenefile);
}

CreateAClone:pitem
{
    SelectItem(pitem.id);
    Clone(1);
    mynewobj=Scene().firstSelect();
    ClearAllBones();

    oldalert=Scene().alertlevel;
	AlertLevel(2);
    SkelegonsToBones();
    AlertLevel(oldalert);
    

    mynewbone=mynewobj.bone();    
    if (!mynewbone)
    {   
        SelectItem(mynewobj.id);
        ClearSelected(); 
        AutoConfirm(0);
        return(nil);
    }
    
	return(mynewobj);
}

PrepareBones:pitem
{
	mybone=pitem.bone();
	i=1;
	while(mybone)
	{
        SelectItem(mybone.id);

 		gbones[i][st_bones]=mybone;
		gbones[i][st_flags]=ResetFlags(mybone);
		gbones[i][st_paren]=mybone.parent;
		gbones[i][st_match]=MatchBone(mybone);
		gbones[i][st_goals]=ResetGoal(mybone);
		gbones[i][st_targe]=ResetTarget(mybone);
		gbones[i][st_mplug]=ResetMotionPlugins(mybone);
//		gbones[i][st_cplug]=ResetChannelPlugins(mybone);
		gbones[i][st_pivot]=ResetPivot(mybone);

		mybone=mybone.next();
		i++;
	}
}

ResetFlags:item
{
	mygoal=item.flags[1];
	if(mygoal)
	{
		BoneActive();
	}
	return(mygoal);
}

MakeNewBoneList:item
{
	mybone=item.bone();
	i=1;
	while(mybone)
	{
		gnewbones[i]=mybone;
		i++;
		mybone=mybone.next();
	}
}

MatchBones
{
	for(i=1;i<=gbones.size();i++)
	{
		myoldbone=gbones[i][st_bones];
		gbones[i][st_match]=MatchBone(myoldbone);
	}
}

MatchBone:oldbone
{

	for(i=1;i<=gnewbones.size();i++)
	{
		mybone=gnewbones[i];
		if(mybone && mybone.name==oldbone.name)
		{
            gnewbones[i]=nil;
            gnewbones.pack();
            gnewbones.trunc();
			return(mybone);
		}
	}
	return(nil);
}

//======================================================================
// Reset Section

ResetGoal:item
{
	if(item.goal)
	{
		mygoal=item.goal;
		GoalItem(0);
		return(mygoal);
	}
	return(nil);
}

ResetTarget:item
{
	if(item.target)
	{
		mygoal=item.target;
		TargetItem(0);
		return(mygoal);
	}
	return(nil);
}

ResetMotionPlugins:item
{
	pluginstate=0;
	i=1;
	while(item.server("ItemMotionHandler",i))
    {
		myserver=item.server("ItemMotionHandler",i);
		pluginstate[i]=item.serverFlags("ItemMotionHandler",i);
		EnableServer("ItemMotionHandler",i,0);
        i++;
    }
    return(pluginstate);
}

ResetChannelPlugins:item
{
	pluginstate=nil;
	i=1;
	while(item.server("ItemChannelHandler",i))
    {
		myserver=item.server("ItemChannelHandler",i);
		pluginstate[i]=item.serverFlags("ItemChannelHandler",i);
		EnableServer("ItemChannelHandler",i,0);
        i++;
    }
    return(pluginstate);
}

ResetPivot:item
{
	mytime=gset[st_fra]/Scene().fps;
    pivrot=item.getPivotRotation(mytime);
    if(pivrot!=<0,0,0>)
    {
        PivotRotation(0,0,0);
        Rotation(pivrot);
        return(1);
    }
    return(nil);
}

//=================================================================

MakeNewBones:olditem,newitem
{
	StatusMsg("Adding new bones");

	SelectItem(olditem.id);
	startpos=gbones.size();

	if(gnewbones[1]==nil)
		return;

	//copy new bones and add them to gbones
	for(i=1;i<=gnewbones.size();i++)
	{
		mybone=gnewbones[i];
		AddBone(mybone.name);
		myoldbone=Scene().firstSelect();

		gbones[startpos+i][st_bones]=myoldbone;
		gbones[startpos+i][st_match]=mybone;
		gbones[startpos+i][st_flags]=gset[st_act];
		gbones[startpos+i][st_pivot]=gset[st_rec];
	}
	// find matching parent
	for(i=1;i<=gnewbones.size();i++)
	{
		mybone=gnewbones[i];
		mypar=mybone.parent;
		myoldpar=FindOldMatch(mypar);
		gbones[startpos+i][st_paren]=myoldpar;
	}
}

PositionBones:pitem
{
	StatusMsg("Positioning bones");
	SetPIP(0);
	d=1;
	for(i=1;i<=gbones.size();i++)
	{
		mysetup=gbones[i];
		myold=mysetup[st_bones];
		mynew=mysetup[st_match];

		if(mynew)
		{
			mynewpar=mynew.parent;
			myoldpar=FindOldMatch(mynewpar);
			SelectItem(myold.id);
			if(myoldpar)
				ParentItem(myoldpar.id);
			else
				ParentItem(pitem.id);

			mypos=mynew.getPosition(0);
			myrot=mynew.getRotation(0);
			myrest=mynew.restlength;
			Position(mypos);
			Rotation(myrot);
//			if(mynew.type==0)   // New for joints
				BoneRestLength(myrest);

		}
		else
		{
			gdelbones[d]=myold;
			gbones[i][1]=nil;
			d++;
		}
	}
	RefreshNow();
}

ReparentBones
{
	StatusMsg("Reparenting bones");

	SetPIP(1);
	for(i=1;i<=gbones.size();i++)
	{
		mybone=gbones[i][st_bones];
		if(mybone)
		{
			currpar=mybone.parent;
			oldpar=gbones[i][st_paren];
			if(oldpar && currpar.id!=oldpar.id)
			{
				SelectItem(mybone.id);
				ParentItem(oldpar.id);
			}
		}
	}
	RefreshNow();
}

DeleteBonesReparent:pitem
{
	StatusMsg("Reparenting children of bones to be deleted");
	if(!gdelbones)
		return;
	SetPIP(1);
	for(i=1;i<=gdelbones.size();i++)
	{
		mybone=gdelbones[i];
		ReparentChildren(mybone);
	}
	
	RefreshNow();
}

ReparentChildren:bone
{
	mychild=bone.firstChild();
	if(!mychild)
		return;
	mypar=FindValidParent(bone);
	mychild=bone.firstChild();
	while(mychild)
	{
		SelectItem(mychild.id);
		ParentItem(mypar.id);
		mychild=bone.firstChild(); // next child becomes first child, so look for that again.
		RefreshNow();
	}
}

FindValidParent:bone
{
	mypar=bone.parent;
	while(mypar)
	{
		found=0;
		for(i=1;i<=gdelbones.size();i++)
		{
			if(gdelbones[i].id==mypar.id)
			{
				mypar=mypar.parent;
				found=1;
				break;
			}
		}
		if(found==0)
			break;
	}
	return(mypar);
}

UnparentBones:plist,pitem
{
	for(i=1;i<=list.size();i++)
	{
		SelectItem(plist[i].id);
		ParentItem(pitem.id);
		RefreshNow();
	}
}

ClearBones:list
{
	StatusMsg("Deleting obsolete bones");

	if(!list)
		return;
	UnparentBones(gdelbones,pitem);
	SelectItem(list[1].id);
	for(i=2;i<=list.size();i++)
	{
		AddToSelection(list[i].id);
	}
	ClearSelected();
	RefreshNow();
}

RestBones
{
	StatusMsg("Resting bones");
	mytime=gset[st_fra]/Scene().fps;
	for(i=1;i<=gbones.size();i++)
	{
		mybone=gbones[i][st_bones];

		if(mybone)
		{
			SelectItem(mybone.id);

			if(gbones[i][st_pivot] && !gbones[i][st_targe])
				myworldrot=SetPivotRotation(mybone);
	
			mypos=mybone.getPosition(mytime);
			myrot=mybone.getRotation(mytime);
			BoneRestPosition(mypos.x,mypos.y,mypos.z);
			BoneRestRotation(myrot.x,myrot.y,myrot.z);
	
			if(gbones[i][st_flags])
				BoneActive();
	
//			if(gbones[i][st_goals])
//				SetGoalItem(mybone,gbones[i][st_goals],myworldrot);


			if(gbones[i][st_targe])
				TargetItem(gbones[i][st_targe].id);
			
			if(gbones[i].size()>=st_mplug)
				my_plugstate=SetPlugins(mybone,gbones[i][st_mplug]);
		}
	}
	RefreshNow();

}

SetPivotRotation:item
{
	mytime=gset[st_fra]/Scene().fps;
    pivrot=item.getRotation(mytime);
    worldpivot=item.getWorldRotation(mytime);
    if(pivrot!=<0,0,0>)
    {
        PivotRotation(pivrot);
        Rotation(<0,0,0>);
        return(worldpivot);
    }
    return(nil);
}

SetPlugins:pitem,pstate
{
	if(!pstate)
		return;
	i=1;
	while(pitem.server("ItemMotionHandler",i))
    {
		EnableServer("ItemMotionHandler",i,!pstate[i]);
        i++;
    }
    return(pluginstate);
}

MoveIKGoals
{
	RefreshNow();
	mytime=gset[st_fra]/Scene().fps;
	index=1;
	SetPIP(1);
	if(gset[st_iku])
	{
		for(i=1;i<=gbones.size();i++)
		{
			if(gbones[i][st_goals])
			{
				mygoal=gbones[i][st_goals];
				mybone=gbones[i][1];
				//remember goal parents
				myparents[index][1]=mygoal;
				myparents[index][2]=mygoal.parent;
				//unparent goal
				SelectItem(mygoal.id);
				if(mygoal.genus==4)
        			ParentItem(FindBoneMesh(mygoal));
				else
					ParentItem(0);
				RefreshNow();
				index++;
				MoveGoalItem(mybone,mygoal);
			}
		}
	}
	for(i=1;i<=gbones.size();i++)
	{
		if(gbones[i][st_goals])
		{
			SelectItem(gbones[i][1].id);
			GoalItem(gbones[i][st_goals].id);
		}
	}
	//reparent goals.
	for(i=1;i<=myparents.size();i++)
	{
		SelectItem(myparents[i][1].id);
		ParentItem(myparents[i][2].id);
		RefreshNow();
	}
}

MoveGoalItem:pitem,pgoal
{
	mytime=gset[st_fra]/Scene().fps;
	mypos=pitem.getWorldPosition(mytime);
	myrot=pitem.getWorldRotation(mytime);
	myprot=pitem.getPivotRotation(mytime);
	if(myprot)
		myrot=prot;

	SelectItem(pgoal.id);

	PivotPosition(mypos);
	PivotRotation(myrot);
	RefreshNow();
	Position(mypos);
	Rotation(<0,0,0>);

	RefreshNow();


//	SelectItem(pitem.id);
//	GoalItem(pgoal.id);
}

FindNewMatch:oldbone
{
	for(i=1;i<=gbones.size();i++)
	{
		temp=gbones[i][st_bones];
		if(temp.id==oldbone.id)
		{
			newbone=gbones[i][st_match];
			return(newbone);
		}
	}
	return(nil);
}

FindOldMatch:newbone
{
	for(i=1;i<=gbones.size();i++)
	{
		temp=gbones[i][st_match];
		if(temp && temp.id==newbone.id)
		{
			oldbone=gbones[i][st_bones];
			return(oldbone);
		}
	}
	return(nil);
}

SetPIP:ison
{
    mypip=Scene().generalopts[3];
    if((mypip && !ison) || (!mypip && ison))
    {
        ParentInPlace();
        RefreshNow();
    }
}

FindBoneMesh:pbone
{
	while(pbone && pbone.genus!=MESH)
	{
		pbone=pbone.parent;
 	}
 	return(pbone);
}

